package java.nio.channels;

import java.nio.channels.spi.AbstractInterruptibleChannel;
import sun.nio.cs.StreamEncoder;
import java.io.Writer;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.Charset;
import sun.nio.cs.StreamDecoder;
import java.io.Reader;
import java.nio.charset.CharsetDecoder;
import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.util.concurrent.ExecutionException;
import java.io.OutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.nio.ByteBuffer;

public final class Channels
{
    private Channels() {
    }
    
    private static void checkNotNull(Object o, String str) {
        if (o == null) {
            throw new NullPointerException("\"" + str + "\" is null!");
        }
        return;
    }
    
    private static void writeFullyImpl(WritableByteChannel writableByteChannel, ByteBuffer byteBuffer) throws IOException {
        while (byteBuffer.remaining() > 0) {
            if (writableByteChannel.write(byteBuffer) <= 0) {
                throw new RuntimeException("no bytes written");
            }
            continue;
        }
    }
    
    private static void writeFully(WritableByteChannel writableByteChannel, ByteBuffer byteBuffer) throws IOException {
        if ((writableByteChannel instanceof SelectableChannel)) {
            SelectableChannel selectableChannel = (SelectableChannel)writableByteChannel;
            synchronized (selectableChannel.blockingLock()) {
                if (!selectableChannel.isBlocking()) {
                    throw new IllegalBlockingModeException();
                }
                writeFullyImpl(writableByteChannel, byteBuffer);
            }
        }
        else {
            writeFullyImpl(writableByteChannel, byteBuffer);
        }
    }
    
    public static InputStream newInputStream(ReadableByteChannel ch) {
        checkNotNull(ch, "ch");
        return new ChannelInputStream(ch);
    }
    
    public static OutputStream newOutputStream(final WritableByteChannel writableByteChannel) {
        checkNotNull(writableByteChannel, "ch");
        return new OutputStream() {
            private ByteBuffer bb = null;
            private byte[] bs = null;
            private byte[] b1 = null;
            
            @Override
            public synchronized void write(int n) throws IOException {
                if (this.b1 == null) {
                    this.b1 = new byte[1];
                }
                this.b1[0] = (byte)n;
                super.write(this.b1);
            }
            
            @Override
            public synchronized void write(byte[] array, int n, int n2) throws IOException {
                if (n < 0 || n > array.length || n2 < 0 || n + n2 > array.length || n + n2 < 0) {
                    throw new IndexOutOfBoundsException();
                }
                if (n2 == 0) {
                    return;
                }
                ByteBuffer bb = (this.bs == array) ? this.bb : ByteBuffer.wrap(array);
                bb.limit(Math.min(n + n2, bb.capacity()));
                bb.position(n);
                this.bb = bb;
                this.bs = array;
                Channels.access$000(writableByteChannel, bb);
                return;
            }
            
            @Override
            public void close() throws IOException {
                writableByteChannel.close();
            }
        };
    }
    
    public static InputStream newInputStream(final AsynchronousByteChannel asynchronousByteChannel) {
        checkNotNull(asynchronousByteChannel, "ch");
        return new InputStream() {
            private ByteBuffer bb = null;
            private byte[] bs = null;
            private byte[] b1 = null;
            
            @Override
            public synchronized int read() throws IOException {
                if (this.b1 == null) {
                    this.b1 = new byte[1];
                }
                if (super.read(this.b1) == 1) {
                    return this.b1[0] & 0xFF;
                }
                return -1;
            }
            
            @Override
            public synchronized int read(byte[] array, int n, int n2) throws IOException {
                if (n < 0 || n > array.length || n2 < 0 || n + n2 > array.length || n + n2 < 0) {
                    throw new IndexOutOfBoundsException();
                }
                if (n2 == 0) {
                    return 0;
                }
                ByteBuffer bb = (this.bs == array) ? this.bb : ByteBuffer.wrap(array);
                bb.position(n);
                bb.limit(Math.min(n + n2, bb.capacity()));
                this.bb = bb;
                this.bs = array;
                int n3 = 0;
                try {
                    return (int)asynchronousByteChannel.read(bb).get();
                }
                catch (ExecutionException ex) {
                    throw new IOException(ex.getCause());
                }
                catch (InterruptedException ex2) {
                    n3 = 1;
                }
                finally {
                    if (n3 != 0) {
                        Thread.currentThread().interrupt();
                    }
                }
            }
            
            @Override
            public void close() throws IOException {
                asynchronousByteChannel.close();
            }
        };
    }
    
    public static OutputStream newOutputStream(final AsynchronousByteChannel asynchronousByteChannel) {
        checkNotNull(asynchronousByteChannel, "ch");
        return new OutputStream() {
            private ByteBuffer bb = null;
            private byte[] bs = null;
            private byte[] b1 = null;
            
            @Override
            public synchronized void write(int n) throws IOException {
                if (this.b1 == null) {
                    this.b1 = new byte[1];
                }
                this.b1[0] = (byte)n;
                super.write(this.b1);
            }
            
            @Override
            public synchronized void write(byte[] array, int n, int n2) throws IOException {
                if (n < 0 || n > array.length || n2 < 0 || n + n2 > array.length || n + n2 < 0) {
                    throw new IndexOutOfBoundsException();
                }
                if (n2 == 0) {
                    return;
                }
                ByteBuffer bb = (this.bs == array) ? this.bb : ByteBuffer.wrap(array);
                bb.limit(Math.min(n + n2, bb.capacity()));
                bb.position(n);
                this.bb = bb;
                this.bs = array;
                boolean n3 = false;
                try {
                    while (bb.remaining() > 0) {
                        try {
                            asynchronousByteChannel.write(bb).get();
                        }
                        catch (ExecutionException ex) {
                            throw new IOException(ex.getCause());
                        }
                        catch (InterruptedException ex2) {
                            n3 = true;
                        }
                    }
                }
                finally {
                    if (!(n3 == 0)) {
                        Thread.currentThread().interrupt();
                    }
                }
                return;
            }
            
            @Override
            public void close() throws IOException {
                asynchronousByteChannel.close();
            }
        };
    }
    
    public static ReadableByteChannel newChannel(InputStream inputStream) {
        checkNotNull(inputStream, "in");
        if (inputStream instanceof FileInputStream && FileInputStream.class.equals(inputStream.getClass())) {
            return ((FileInputStream)inputStream).getChannel();
        }
        return new ReadableByteChannelImpl(inputStream);
    }
    
    public static WritableByteChannel newChannel(OutputStream outputStream) {
        checkNotNull(outputStream, "out");
        if (outputStream instanceof FileOutputStream && FileOutputStream.class.equals(outputStream.getClass())) {
            return ((FileOutputStream)outputStream).getChannel();
        }
        return new WritableByteChannelImpl(outputStream);
    }
    
    public static Reader newReader(ReadableByteChannel ch, CharsetDecoder charsetDecoder, int minBufferCap) {
        checkNotNull(ch, "ch");
        return StreamDecoder.forDecoder(ch, charsetDecoder.reset(), minBufferCap);
    }
    
    public static Reader newReader(ReadableByteChannel readableByteChannel, String charsetName) {
        checkNotNull(charsetName, "csName");
        return newReader(readableByteChannel, Charset.forName(charsetName).newDecoder(), -1);
    }
    
    public static Writer newWriter(WritableByteChannel writableByteChannel, CharsetEncoder charsetEncoder, int n) {
        checkNotNull(writableByteChannel, "ch");
        return StreamEncoder.forEncoder(writableByteChannel, charsetEncoder.reset(), n);
    }
    
    public static Writer newWriter(WritableByteChannel writableByteChannel, String charsetName) {
        checkNotNull(charsetName, "csName");
        return newWriter(writableByteChannel, Charset.forName(charsetName).newEncoder(), -1);
    }
    
    static void access$000(WritableByteChannel writableByteChannel, ByteBuffer byteBuffer) throws IOException {
        writeFully(writableByteChannel, byteBuffer);
    }
    
    private static class ReadableByteChannelImpl extends AbstractInterruptibleChannel implements ReadableByteChannel
    {
        InputStream in;
        private static final int TRANSFER_SIZE = 8192;
        private byte[] buf;
        private boolean open;
        private Object readLock;
        
        ReadableByteChannelImpl(InputStream in) {
            this.buf = new byte[0];
            this.open = true;
            this.readLock = new Object();
            this.in = in;
        }
        
        @Override
        public int read(ByteBuffer byteBuffer) throws IOException {
            int remaining = byteBuffer.remaining();
            int i = 0;
            int read = 0;
            synchronized (this.readLock) {
                while (i < remaining) {
                    int min = Math.min(remaining - i, 8192);
                    if (this.buf.length < min) {
                        this.buf = new byte[min];
                    }
                    if (i > 0 && this.in.available() <= 0) {
                        break;
                    }
                    try {
                        this.begin();
                        read = this.in.read(this.buf, 0, min);
                    }
                    finally {
                        this.end(read > 0);
                    }
                    if (read < 0) {
                        break;
                    }
                    i += read;
                    byteBuffer.put(this.buf, 0, read);
                }
                if (read < 0 && i == 0) {
                    return -1;
                }
                return i;
            }
        }
        
        @Override
        protected void implCloseChannel() throws IOException {
            this.in.close();
            this.open = false;
        }
    }
    
    private static class WritableByteChannelImpl extends AbstractInterruptibleChannel implements WritableByteChannel
    {
        OutputStream out;
        private static final int TRANSFER_SIZE = 8192;
        private byte[] buf;
        private boolean open;
        private Object writeLock;
        
        WritableByteChannelImpl(OutputStream out) {
            this.buf = new byte[0];
            this.open = true;
            this.writeLock = new Object();
            this.out = out;
        }
        
        @Override
        public int write(ByteBuffer byteBuffer) throws IOException {
            int remaining = byteBuffer.remaining();
            int i = 0;
            synchronized (this.writeLock) {
                while (i < remaining) {
                    int min = Math.min(remaining - i, 8192);
                    if (this.buf.length < min) {
                        this.buf = new byte[min];
                    }
                    byteBuffer.get(this.buf, 0, min);
                    try {
                        this.begin();
                        this.out.write(this.buf, 0, min);
                    }
                    finally {
                        this.end(min > 0);
                    }
                    i += min;
                }
                return i;
            }
        }
        
        @Override
        protected void implCloseChannel() throws IOException {
            this.out.close();
            this.open = false;
        }
    }
}
